/********************************************************************
 *                                                                  *
 *                             LINACQ                               *
 *                            v0.12beta                             *
 *                                                                  *
 *                                                                  *
 *   LINux ACQuisition program.                                     *
 *   Interfacer med et Computer boards DAS1602/16, multifunction    *
 *   I/O board.                                                     *
 *                                                                  *
 *                                                                  *
 *   Copyright istein Aanensen, IFJF 2000                          *
 *                                                                  *
 ********************************************************************/
#define _GNU_SOURCE
#include <getopt.h>
#include <sys/ioctl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <pthread.h>
#include "segy.h"
#include "pci-das1602.h"

#define VERSION "v0.12beta"
#define MODEMDEVICE "/dev/ttyS0"
char *gps_time;
int east, north, event;

void printHelp();
void readEiva(int fd);
void initializeBoard(int fd, unsigned short pace_freq, unsigned short nch);

int main(int argc, char *argv[]) {
  char *filename;
  char qc_char='0';
  short do_qc=0, do_eiva=0;
  int qc_fd[2];
  int su_fd, fd, serial_fd, fifo_fd,qc_buf=0, bytes, pace_freq=100000, count=2048;
  int opt, opt_ind;
  unsigned int no_shots=0, i=0, j=0, ffid=1, nch=4;
  unsigned short (*data)[16*32764];
  pthread_t eiva_thread;
  segy tr;
  
  static struct option arg_opt[] =
  {
    {"file",1,0,'f'},
    {"nshots",1,0,'n'},
    {"ns",2,0,'s'},
    {"nch",2,0,'c'},
    {"freq",2,0,'t'},
    {"eiva",0,0,'e'},
    {"qc",0,0,'q'},
    {"help",0,0,'h'}
  };
  
  /* read program arguments */
  while ((opt=getopt_long_only(argc, argv, "hqec:s:t:f:n:", 
			  arg_opt, &opt_ind)) != -1) {
    switch (opt) {
    case 'q':
      do_qc=1;
      break;
    case 'e':
      do_eiva=1;
      break;
    case 's':
       count=atoi(optarg);
       break;
    case 'c':
      nch=atoi(optarg);
      break;
    case 'n':
      no_shots=atoi(optarg);
      break;
    case 't':
      pace_freq=atoi(optarg);
      break;
    case 'f':
      filename=optarg;
      break;
    case 'h':
      printHelp();
      exit(0);
    }
  }

  if ( (filename == NULL) || (no_shots == 0)) {
    printf("Must specify -file= -nshots=\n");
    exit(-1);
  }
  count *= nch;
  printf("Total no samples: %i\n",count);
  
  /* Dynamic allocation */
  data = malloc(count*sizeof(data));
  if (data == NULL)
    exit(-1);
  printf("%i kbytes allocated in buffer.\n",count*sizeof(data)/1024);
  
  fd = open("/dev/pci-das1602-16_adc00", ADC_PACER_CLOCK);                 /* AD board device  */
  if (fd <0) {perror("/dev/pci-das1602-16_adc00"); exit(-1); }
  initializeBoard(fd, pace_freq,nch);                                      /* Initialize board */

  su_fd=open(filename,O_CREAT | O_WRONLY | O_TRUNC, S_IREAD | S_IWRITE);
  if (su_fd <0) {perror(argv[1]); exit(-1); }

  if (do_qc) {
    qc_fd[0]=open("/qc3/qc/data.su0", O_CREAT | O_WRONLY | O_TRUNC); /* qc_file buffer 0 */
    if (qc_fd[0] <0) {perror("/tmp/data.su0"); exit(-1); }
    qc_fd[1]=open("/qc3/qc/data.su1", O_CREAT | O_WRONLY | O_TRUNC); /* qc_file buffer 1 */
    if (qc_fd[1] <0) {perror("/tmp/data.su1"); exit(-1); }
    printf("Waiting for the QC kernel to start...\n");
    fifo_fd=open("/qc3/qc/qc_trig", O_WRONLY);                       /* qc fifo trigger  */
    if (fifo_fd <0) {perror("/tmp/qc_trig"); exit(-1); }
  }

  if (do_eiva) {
    serial_fd = open(MODEMDEVICE, O_RDWR | O_NOCTTY );     /* serial port, com. with eiva */
    if (serial_fd <0) {perror(MODEMDEVICE); exit(-1); }
    printf("Connecting to Eiva nav......\n");
    pthread_create(&eiva_thread, NULL, (void*) readEiva, &serial_fd);
    sleep(1);

    printf("Connection established...\n");
  }
  printf("Linacq kernel started, waiting for trigger...\n");
  printf("%s open for output, %i shots for %i channels.\n",filename,no_shots,nch);  
  
  while (ffid <= no_shots) {
    bytes=read(fd, data, count);                /* Read count samples of data         */
    
    if (bytes == count) {                   /* Write to file, if all samples only */
      tr.dt=1000000/pace_freq;                  /* Sample rate in us (micro sec.)      */
      tr.fldr=ffid;
      tr.ns=count/nch;
      if (do_eiva) {
	tr.ep=event;
	tr.sx=east;
	tr.sy=north;
      }

      /* Demultiplex data */
      for (j=0;j<nch;j++) {
	for (i=0;i<count;i=i+nch) 
	  tr.data[i/nch]=(float) (((*data)[i+j]-32768)/3276.8);
      
	/* Start the write procedure */
	tr.tracf=j+1;
	tr.tracl=ffid;
	tr.tracr=j+1;
	/* Writes the trace header 240bytes + 4bytes*no_samples, (sizeof(float)=4)     */
	bytes=write(su_fd, &tr, 240+(4*count/nch));
	printf("%i bytes written. Cable no. %i.\n",bytes,j+1);
	if (do_qc)
	  bytes=write(qc_fd[qc_buf], &tr, 240+(4*count/nch));    /* wites to qc buffer */
      }
      if (do_eiva)
	printf("%s %i %i %i ",gps_time, east, north, event);
      printf("%i\n",ffid);
      
      if (do_qc) {
	write(fifo_fd, &qc_char, 1);                             /* qc fifo trigger char */
	/* Handlig the two qc files */
	lseek(qc_fd[qc_buf], 0, SEEK_SET);                       /* Overwrite qc file each time */

	if (qc_buf == 1) {
	  qc_buf=0;
	  qc_char='0';
	} else {
	  qc_buf++;
	  qc_char='1';            /*  qc buffers flip flop */
	}
      }
      
    } else {
      fputc('\a', stdout);                       /* Makes a beep in the pc speaker */
      printf("Error: couldn't read correct no of samples\n");
      printf("FFID: %i lost!\n",ffid);
      printf("Got only %i bytes\n", bytes);
    }
    ffid++;
  }
  
  close(fd);
  close(su_fd);
  close(serial_fd);
  close(fifo_fd);
  close(qc_fd[0]);
  close(qc_fd[1]);
  exit(0);
}

void initializeBoard(int fd, unsigned short pace_freq, unsigned short nch) {
  short data, bytes;
  ioctl(fd, ADC_SET_FRONT_END, 1);              /* 1=singel ended, 0=diffrential mode */
  ioctl(fd, ADC_SET_PACER_FREQ, pace_freq);     /* Sample rate in hertz               */
  ioctl(fd, ADC_SET_TRIGGER, TRIG_POS_EDGE);    /* Trig method                        */
  ioctl(fd, ADC_SET_MUX_LOW, 0);                /* First channel to scan              */
  ioctl(fd, ADC_SET_MUX_HIGH, nch-1);           /* Last channel to scan               */
  ioctl(fd, ADC_BURST_MODE, 1);                 /* 1=burst mode on, 0=burst mode off  */
  ioctl(fd, ADC_SET_GAINS, BP_10_00V);          /* Programmeble gain control          */
  bytes=read(fd, &data, 1);                     /* Initsialize board                  */
}

void readEiva(int fd) {
  char eiva_string[256];
  char *bytes;
  FILE *eiva_fd;
  /* Setting the serial port to read raw data */
  struct termios options;
  tcgetattr(fd,&options);      /* Get current setting */
  options.c_cflag &= ~PARENB;
  options.c_cflag &= ~CSTOPB;
  options.c_cflag &= ~CSIZE; /* Mask the character size bits */
  options.c_cflag |= CS8;    /* Select 8 data bits */
  options.c_lflag |= (ICANON | ECHO | ECHOE);
  tcsetattr(fd, TCSANOW, &options); /* Apply changes */
  /* Finished serial settings */
  close(fd);
  
  eiva_fd = fopen(MODEMDEVICE,"r");
  while (1) {
    bytes=fgets(eiva_string,sizeof eiva_string, eiva_fd);
    /* Exstract the correct eiva format, time east north event */
    gps_time=strtok(eiva_string," ");
    east=atoi(strtok(NULL," "));
    north=atoi(strtok(NULL," "));
    event=atoi(strtok(NULL," "));
  }
}

void printHelp () {
  printf(" Legal options are:\n");
  printf("\n");
  printf("   -file=<filename>          specify SU output file\n");
  printf("   -nshots=<no shots>        number of shots, before linacq exits\n");
  printf("   -ns=<no samples/ch>       number of samples per channel\n");
  printf("   -nch=<no channels>        number of channels (max 8)\n");
  printf("   -freq=<frequency>         sampling frequency\n");
  printf("   -eiva                     connect to eiva navigation\n");
  printf("   -qc                       connect to QCIX QC system\n");
  printf("   -help                     print this message\n");
  printf("\n");
  printf("\n");
  printf("             LINACQ %s\n",VERSION);
  printf(" Linux acquisition system, by istein Aanensen\n");
  printf("\n");
}
